/*	$NetBSD$	*/

/*-
 * Copyright (c) 2009 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * STR91XX/STR92XX Orion series SoC
 * Embedded 4-port Gigabit Switch Engine with two 10/100/1000M RGMII/MII ports
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#define DEBUG_GSE_DUMP
#define DEBUG_GSE

#include "bpfilter.h"
#include "vlan.h"
#include "rnd.h"

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/mbuf.h>
#include <sys/device.h>
#include <sys/sockio.h>
#include <sys/kernel.h>
#include <sys/callout.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_ether.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif
#if NRND > 0
#include <sys/rnd.h>
#endif

#include <dev/mii/mii.h>
#include <dev/mii/miivar.h>

#include <arm/star/starvar.h>
#include <arm/star/if_gsereg.h>

#define GSE_ETHER_MIN_LEN	(ETHER_MIN_LEN - ETHER_CRC_LEN)	/* not including FCS */
#define GSE_ETHER_MIN_LEN_VLAN	(GSE_ETHER_MIN_LEN + ETHER_VLAN_ENCAP_LEN)	/* not including VLAN-tag and FCS */
#define GSE_MAX_PKT_LEN		1536
#define GSE_MAX_PKT_NSEGS	16	/* XXX */

#define GSE_TX_RING_CNT		256	/* must be 2^n */
#define GSE_TX_RING_MINIMAL_CNT	1	/* ready to build padding */
#define GSE_RX_RING_CNT		512	/* must be 2^n */

#define GSE_TX_NEXTIDX(idx)	(((idx) + 1) & (GSE_TX_RING_CNT - 1))
#define GSE_RX_NEXTIDX(idx)	(((idx) + 1) & (GSE_RX_RING_CNT - 1))


struct gse_softc {
	device_t sc_dev;

	bus_addr_t sc_addr;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	bus_dma_tag_t sc_dmat;

	/* interrupts */
	void *sc_ih_mib;	/* IRQ18 MIB counter, Port status changed */
	void *sc_ih_tx;		/* IRQ19 TX complete */
	void *sc_ih_rx;		/* IRQ20 RX complete */
	/* IRQ21 (Queue Empty) and IRQ22 (Queue Full) are not used */

	/* TX */
	struct gse_txdesc *sc_txdesc_ring;	/* gse_txdesc[GSE_TX_RING_CNT] */
	bus_dmamap_t sc_txdesc_dmamap;
	struct gse_rxdesc *sc_rxdesc_ring;	/* gse_rxdesc[GSE_RX_RING_CNT] */
	bus_dmamap_t sc_rxdesc_dmamap;
	void *sc_padseg;			/* 64byte stub segment */
	bus_dmamap_t sc_padseg_dmamap;		/* pad for short TX frame */
	struct gse_txsoft {
		struct mbuf *txs_mbuf;		/* head of our mbuf chain */
		bus_dmamap_t txs_dmamap;	/* our DMA map */
	} sc_txsoft[GSE_TX_RING_CNT];
	int sc_tx_considx;
	int sc_tx_prodidx;
	int sc_tx_free;

	/* RX */
	struct gse_rxsoft {
		struct mbuf *rxs_mbuf;		/* head of our mbuf chain */
		bus_dmamap_t rxs_dmamap;	/* our DMA map */
	} sc_rxsoft[GSE_RX_RING_CNT];
	int sc_rx_readidx;
	uint32_t sc_rx_last_dptr;
	callout_t sc_rx_ch;			/* RX stall check callout */

	/* misc */
	int sc_if_flags;			/* local copy of if_flags */
	int sc_flowflags;			/* 802.3x flow control flags */
	struct ethercom sc_ethercom;		/* interface info */
	struct mii_data sc_mii;
	uint8_t sc_enaddr[ETHER_ADDR_LEN];
#if NRND > 0
	rndsource_element_t sc_rnd_source;
#endif
};

#define TXDESC_WRITEOUT(idx)					\
	bus_dmamap_sync(sc->sc_dmat, sc->sc_txdesc_dmamap,	\
	    sizeof(struct gse_txdesc) * (idx),			\
	    sizeof(struct gse_txdesc),				\
	    BUS_DMASYNC_PREWRITE)

#define TXDESC_READIN(idx)					\
	bus_dmamap_sync(sc->sc_dmat, sc->sc_txdesc_dmamap,	\
	    sizeof(struct gse_txdesc) * (idx),			\
	    sizeof(struct gse_txdesc),				\
	    BUS_DMASYNC_PREREAD)

#define RXDESC_WRITEOUT(idx)					\
	bus_dmamap_sync(sc->sc_dmat, sc->sc_rxdesc_dmamap,	\
	    sizeof(struct gse_rxdesc) * (idx),			\
	    sizeof(struct gse_rxdesc),				\
	    BUS_DMASYNC_PREWRITE)

#define RXDESC_READIN(idx)					\
	bus_dmamap_sync(sc->sc_dmat, sc->sc_rxdesc_dmamap,	\
	    sizeof(struct gse_rxdesc) * (idx),			\
	    sizeof(struct gse_rxdesc),				\
	    BUS_DMASYNC_PREREAD)

#define GSE_REG_SETBIT(sc, reg, bit)					\
	bus_space_write_4((sc)->sc_iot, (sc)->sc_ioh, reg,		\
	    bus_space_read_4((sc)->sc_iot, (sc)->sc_ioh, reg) | (bit))

#define GSE_REG_CLRBIT(sc, reg, bit)					\
	bus_space_write_4((sc)->sc_iot, (sc)->sc_ioh, reg,		\
	    bus_space_read_4((sc)->sc_iot, (sc)->sc_ioh, reg) & ~(bit))

#define TX_DPTR2IDX(dptr)					\
	(((dptr) - sc->sc_txdesc_dmamap->dm_segs[0].ds_addr) /	\
	sizeof(struct gse_txdesc))

#define RX_DPTR2IDX(dptr)					\
	(((dptr) - sc->sc_rxdesc_dmamap->dm_segs[0].ds_addr) /	\
	sizeof(struct gse_rxdesc))


static int gse_match(device_t, struct cfdata *, void *);
static void gse_attach(device_t, device_t, void *);

/* interrupt handlers */
static int gse_err_intr(void *);
static int gse_tx_intr(void *);
static int gse_rx_intr(void *);

/* for ifnet interface */
static void gse_start(struct ifnet *);
static int gse_ioctl(struct ifnet *, u_long, void *);
static int gse_init(struct ifnet *);
static void gse_stop(struct ifnet *, int);
static void gse_watchdog(struct ifnet *);

/* for MII */
static int gse_miibus_readreg(device_t, int, int);
static void gse_miibus_writereg(device_t, int, int, int);
static void gse_miibus_statchg(device_t);

/* misc routines */
static void gse_setmulti(struct gse_softc *);
static void gse_clear_hashtable(struct gse_softc *);
static void gse_add_hashtable(struct gse_softc *, uint8_t *);
static void gse_apply_hashtable(struct gse_softc *);
static void gse_stall_recover(void *arg);
static int gse_stall_check(struct gse_softc *, int);
static int gse_encap(struct gse_softc *, struct mbuf *);
static int gse_init_regs(struct gse_softc *);
static int gse_alloc_ring(struct gse_softc *);
static void gse_init_txring(struct gse_softc *);
static int gse_init_rxring(struct gse_softc *);
static void gse_reset_rxdesc(struct gse_softc *, int);
static int gse_alloc_rxbuf(struct gse_softc *, int);
static void gse_drain_txbuf(struct gse_softc *sc);
static void gse_drain_rxbuf(struct gse_softc *sc);
static int gse_alloc_dma(struct gse_softc *, size_t, void **,
                         bus_dmamap_t *);

#ifdef DEBUG_GSE
static void gse_dump_txdesc(struct gse_softc *, int, const char *);
static void gse_dump_rxdesc(struct gse_softc *, int, const char *);
#endif

CFATTACH_DECL_NEW(gse, sizeof(struct gse_softc),
    gse_match, gse_attach, NULL, NULL);

/* ARGSUSED */
static int
gse_match(device_t parent __unused, struct cfdata *match __unused, void *aux)
{
	struct star_attach_args *sa;

	if (!CPU_IS_STR9100())
		return 0;

	sa = (struct star_attach_args *)aux;

	sa->sa_size = GSE_REG_SIZE;
	return 1;
}

/* ARGSUSED */
static void
gse_attach(device_t parent __unused, device_t self, void *aux)
{
	struct gse_softc *sc;
	struct star_attach_args *sa;
	struct ifnet *ifp;

	sa = aux;
	sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = sa->sa_iot;
	sc->sc_addr = sa->sa_addr;
	sc->sc_dmat = sa->sa_dmat;

	aprint_naive("\n");
	aprint_normal(": Gigabit Switch Engine\n");

	if (bus_space_map(sc->sc_iot, sc->sc_addr, sa->sa_size, 0,
	    &sc->sc_ioh)) {
		aprint_error_dev(self, "Cannot map registers\n");
		return;
	}

	/* allocate dma buffer */
	if (gse_alloc_ring(sc))
		return;
	gse_init_regs(sc);


	/* setup my mac address */
	// notyet

	/*
	 * setup interrupt handlers
	 */
	if ((sc->sc_ih_mib = star_intr_establish(STAR_IRQ_NIC_STAT, IPL_NET,
	    gse_err_intr, sc)) == NULL) {
		aprint_error_dev(self, "unable to establish MIB interrupt\n");
		goto failure;
	}
	if ((sc->sc_ih_tx = star_intr_establish(STAR_IRQ_NIC_TX, IPL_NET,
	    gse_tx_intr, sc)) == NULL) {
		aprint_error_dev(self, "unable to establish TX interrupt\n");
		goto failure;
	}
	if ((sc->sc_ih_rx = star_intr_establish(STAR_IRQ_NIC_RX, IPL_NET,
	    gse_rx_intr, sc)) == NULL) {
		aprint_error_dev(self, "unable to establish RX interrupt\n");
		goto failure;
	}
	callout_init(&sc->sc_rx_ch, 0);
	callout_setfunc(&sc->sc_rx_ch, gse_stall_recover, sc);

	/*
	 * setup ifp
	 */
	ifp = &sc->sc_ethercom.ec_if;
	strlcpy(ifp->if_xname, device_xname(sc->sc_dev), IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_mtu = ETHERMTU;
	ifp->if_baudrate = IF_Mbps(1000);
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = gse_ioctl;
	ifp->if_start = gse_start;
	ifp->if_init = gse_init;
	ifp->if_stop = gse_stop;
	ifp->if_watchdog = gse_watchdog;

#if 0
	ifp->if_capabilities =
	    IFCAP_CSUM_IPv4_Tx |
	    IFCAP_CSUM_IPv4_Rx |
	    IFCAP_CSUM_TCPv4_Tx | IFCAP_CSUM_TCPv4_Rx |
	    IFCAP_CSUM_UDPv4_Tx | IFCAP_CSUM_UDPv4_Rx;
#endif

	IFQ_SET_MAXLEN(&ifp->if_snd, max(GSE_TX_RING_CNT - 1, IFQ_MAXLEN));
	IFQ_SET_READY(&ifp->if_snd);

	/* setup MII */
	sc->sc_ethercom.ec_mii = &sc->sc_mii;
	sc->sc_mii.mii_ifp = ifp;
	sc->sc_mii.mii_readreg = gse_miibus_readreg;
	sc->sc_mii.mii_writereg = gse_miibus_writereg;
	sc->sc_mii.mii_statchg = gse_miibus_statchg;

	ifmedia_init(&sc->sc_mii.mii_media, 0, ether_mediachange,
	    ether_mediastatus);

	/* attach PHY */
	mii_attach(self, &sc->sc_mii, 0xffffffff, MII_PHY_ANY,
	    MII_OFFSET_ANY, 0);

	if (LIST_FIRST(&sc->sc_mii.mii_phys) == NULL) {
		aprint_error_dev(self, "no PHY found on MII\n");
		ifmedia_add(&sc->sc_mii.mii_media, IFM_ETHER | IFM_MANUAL,
		    0, NULL);
		ifmedia_set(&sc->sc_mii.mii_media, IFM_ETHER | IFM_MANUAL);
	} else {
		ifmedia_set(&sc->sc_mii.mii_media, IFM_ETHER | IFM_AUTO);
	}

	if_attach(ifp);
	ether_ifattach(ifp, sc->sc_enaddr);

#ifdef notyet
	if (!pmf_device_register(self, NULL, gse_resume))
		aprint_error_dev(self, "couldn't establish power handler\n");
	else
		pmf_class_network_register(self, ifp);
#endif

#if NRND > 0
	rnd_attach_source(&sc->sc_rnd_source, device_xname(sc->sc_dev),
	    RND_TYPE_NET, 0);
#endif
	return;

 failure:
	return;
}

static int
gse_err_intr(void *arg)
{
	struct gse_softc *sc = arg;
	uint32_t status;

	status = bus_space_read_4(sc->sc_iot, sc->sc_ioh, GSE_INT_STATUS_REG);

#ifdef GSE_DEBUG
	if (status) {
		char tmpbuf[128];
		snprintb(tmpbuf, sizeof(tmpbuf), 
		    "\20"
		    "\x20PORT1_INGRESS"
		    "\x1fPORT1_LOCAL"
		    "\x1ePORT1_RMC_PAUSE"
		    "\x1dPORT1_NO_DEST"
		    "\x1cPORT1_JAMMED"
		    "\x1bPORT1_RX_ERR"
		    "\x1aPORT1_BCSTORM"
		    "\x19PORT1_NOFREELINK"
		    "\x18PORT0_INGRESS"
		    "\x17PORT0_LOCAL"
		    "\x16PORT0_RMC_PAUSE"
		    "\x15PORT0_NO_DEST"
		    "\x14PORT0_JAMMED"
		    "\x13PORT0_RX_ERR"
		    "\x12PORT0_BCSTORM"
		    "\x11PORT0_NOFREELINK"
		    "\x0cCPU_UNKNOWN_VLAN"
		    "\x0bPORT1_UNKNOWN_VLAN"
		    "\x0aPORT0_UNKNOWN_VLAN"
		    "\x09INTRUDER1"
		    "\x08INTRUDER0"
		    "\x07PORT_STATUS_CHG"
		    "\x06BUFFER_FULL"
		    "\x05GLOBAL_Q_FULL"
		    "\x04HNAT_Q_FULL"
		    "\x03CPU_Q_FULL"
		    "\x02PORT1_Q_FULL"
		    "\x01PORT0_Q_FULL"
		    status);
		printf("%s:%d: sc=%p, status=%s\n", __func__, __LINE__, sc, tmpbuf);
	}
#endif

	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_INT_STATUS_REG, status);
	return 1;
}

static int
gse_tx_intr(void *arg)
{
	struct gse_softc *sc;
	struct ifnet *ifp;
	struct gse_txsoft *txs;
	int idx;

	sc = (struct gse_softc *)arg;
	ifp = &sc->sc_ethercom.ec_if;

	for (idx = sc->sc_tx_considx; idx != sc->sc_tx_prodidx;
	    idx = GSE_TX_NEXTIDX(idx)) {

		txs = &sc->sc_txsoft[idx];

#ifdef DEBUG_GSE_DUMP
		gse_dump_txdesc(sc, idx, "TX_INTR:");
#endif

		TXDESC_READIN(idx);
		if (!(sc->sc_txdesc_ring[idx].tx_ctrl & GSE_TXDESC_CTRL_COWN)) {
			/* This TX Descriptor has not been transmitted yet? */
			break;
		}

		/* txsoft is used only first segment */
		if (sc->sc_txdesc_ring[idx].tx_ctrl & GSE_TXDESC_CTRL_FS) {
			bus_dmamap_unload(sc->sc_dmat,
			    txs->txs_dmamap);
			m_freem(txs->txs_mbuf);

			ifp->if_opackets++;
		}
		sc->sc_tx_free++;
	}
	sc->sc_tx_considx = idx;

#ifdef DEBUG_GSE_DUMP
	printf("TX_INTR:tx_free=%d\n", sc->sc_tx_free);
#endif

	if (sc->sc_tx_free > GSE_TX_RING_MINIMAL_CNT)
		ifp->if_flags &= ~IFF_OACTIVE;

	/*
	 * No more pending TX descriptor,
	 * cancel the watchdog timer.
	 */
	if (sc->sc_tx_free == GSE_TX_RING_CNT)
		ifp->if_timer = 0;

	return 1;
}

static int
gse_rx_intr(void *arg)
{
	struct gse_softc *sc;
	struct ifnet *ifp;
	struct gse_rxsoft *rxs;
	int idx, len;
	uint32_t ctrl;
	struct mbuf *m;

	sc = (struct gse_softc *)arg;
	ifp = &sc->sc_ethercom.ec_if;

	for (idx = sc->sc_rx_readidx; RX_DPTR2IDX(bus_space_read_4(sc->sc_iot,
	    sc->sc_ioh, GSE_RX_DPTR_REG)) != idx; idx = GSE_RX_NEXTIDX(idx)) {

		rxs = &sc->sc_rxsoft[idx];

		RXDESC_READIN(idx);
		ctrl = sc->sc_rxdesc_ring[idx].rx_ctrl;
		if ((ctrl & GSE_RXDESC_CTRL_COWN) == 0)
			continue;

		/*
		 * build mbuf from RX Descriptor if needed
		 */
		m = rxs->rxs_mbuf;
		if (m == NULL) {
			/* mbuf empty? allocate new mbuf cluster */
			gse_alloc_rxbuf(sc, idx);
			m = rxs->rxs_mbuf;
			if (m == NULL) {
				printf("%s: mbuf allocation failed."
				    " dropping packet. FATAL ERROR!\n",
				    device_xname(sc->sc_dev));
				continue;
			}
		}

		len = ctrl & GSE_RXDESC_CTRL_SDL_MASK;
		if (len < ETHER_HDR_LEN) {
			ifp->if_ierrors++;
			goto rx_failure;
		}

		/* packet receive ok */
		ifp->if_ipackets++;
		m->m_pkthdr.rcvif = ifp;
		m->m_pkthdr.len = m->m_len = len;

		/* checksum offloading */
		if (ifp->if_csum_flags_rx & M_CSUM_IPv4) {
			if ((ctrl & GSE_RXDESC_CTRL_PROT_MASK) !=
			    GSE_RXDESC_CTRL_PROT_OTHERS) {
				/* IP packet (including TCP and UDP) */
				m->m_pkthdr.csum_flags |= M_CSUM_IPv4;
				if (ctrl & GSE_RXDESC_CTRL_IPF)
					m->m_pkthdr.csum_flags |= M_CSUM_IPv4_BAD;
			}
		}
		if (ifp->if_csum_flags_rx & M_CSUM_TCPv4) {
			if ((ctrl & GSE_RXDESC_CTRL_PROT_MASK) ==
			    GSE_RXDESC_CTRL_PROT_TCP) {
				m->m_pkthdr.csum_flags |= M_CSUM_TCPv4;
				if (ctrl & GSE_RXDESC_CTRL_L4F)
					m->m_pkthdr.csum_flags |= M_CSUM_TCP_UDP_BAD;
			}
		}
		if (ifp->if_csum_flags_rx & M_CSUM_UDPv4) {
			if ((ctrl & GSE_RXDESC_CTRL_PROT_MASK) ==
			    GSE_RXDESC_CTRL_PROT_UDP) {
				m->m_pkthdr.csum_flags |= M_CSUM_UDPv4;
				if (ctrl & GSE_RXDESC_CTRL_L4F)
					m->m_pkthdr.csum_flags |= M_CSUM_TCP_UDP_BAD;
			}
		}

		bus_dmamap_sync(sc->sc_dmat, rxs->rxs_dmamap, 0,
		    rxs->rxs_dmamap->dm_mapsize, BUS_DMASYNC_PREREAD);

#if NBPFILTER > 0
		/* Pass this up to any BPF listeners. */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif /* NBPFILTER > 0 */

		(*ifp->if_input)(ifp, m);

		/* clear this rxsoft */
		bus_dmamap_unload(sc->sc_dmat, rxs->rxs_dmamap);
		rxs->rxs_mbuf = NULL;

		/* allocate new mbuf cluster or reuse old mbuf*/
		gse_alloc_rxbuf(sc, idx);

	 rx_failure:
		gse_reset_rxdesc(sc, idx);
	}
	sc->sc_rx_readidx = idx;

	if (gse_stall_check(sc, idx)) {
		/* possibility of stalling... */
		callout_schedule(&sc->sc_rx_ch, hz / 1000);
	}

	/* re-enable RX DMA */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_DMA_CTRL_REG,
	    GSE_RX_DMA_CTRL_RX_EN);
	return 1;
}

static void
gse_stall_recover(void *arg)
{
	struct gse_softc *sc;

	sc = (struct gse_softc *)arg;

	if (gse_stall_check(sc, sc->sc_rx_readidx)) {
		/* detect stalling... */
		gse_reset_rxdesc(sc, sc->sc_rx_readidx);

		/* re-enable RX */
		bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_DMA_CTRL_REG,
		    1);

		printf("%s: RX Ring stalled. recovered\n", device_xname(sc->sc_dev));
	}
}

static int
gse_stall_check(struct gse_softc *sc, int idx)
{
	if (RX_DPTR2IDX(bus_space_read_4(sc->sc_iot, sc->sc_ioh,
	    GSE_RX_DPTR_REG)) == idx) {

		RXDESC_READIN(idx);
		if (sc->sc_rxdesc_ring[idx].rx_ctrl & GSE_RXDESC_CTRL_COWN) {
			return 1;
		}
	}
	return 0;
}

static void
gse_clear_hashtable(struct gse_softc *sc)
{
	// XXX: notyet
}

static void
gse_add_hashtable(struct gse_softc *sc, uint8_t *eaddr)
{
	// XXX: notyet
}

static void
gse_apply_hashtable(struct gse_softc *sc)
{
	// XXX: notyet
}

static void
gse_setmulti(struct gse_softc *sc)
{
	struct ifnet *ifp;
	struct ether_multi *enm;
	struct ether_multistep step;

	ifp = &sc->sc_ethercom.ec_if;
	if (ifp->if_flags & IFF_PROMISC) {
 allmulti:
		ifp->if_flags |= IFF_ALLMULTI;
		return;
	}

	gse_clear_hashtable(sc);
	ifp->if_flags &= ~IFF_ALLMULTI;

	ETHER_FIRST_MULTI(step, &sc->sc_ethercom, enm);
	while (enm != NULL) {
		/*
		 * If multicast range, fall back to ALLMULTI.
		 */
		if (memcmp(enm->enm_addrlo, enm->enm_addrhi,
		    ETHER_ADDR_LEN) != 0)
			goto allmulti;

		gse_add_hashtable(sc, enm->enm_addrlo);
		ETHER_NEXT_MULTI(step, enm);
	}

	gse_apply_hashtable(sc);
}

/*
 * ifnet interfaces
 */
static int
gse_init(struct ifnet *ifp)
{
	struct gse_softc *sc;
	int s, error = 0;

	sc = ifp->if_softc;

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d: ifp=%p\n", __func__, __LINE__, ifp);
#endif

	s = splnet();

	gse_init_regs(sc);
	gse_init_txring(sc);
	error = gse_init_rxring(sc);
	if (error != 0) {
		gse_drain_rxbuf(sc);
		printf("%s: Cannot allocate mbuf cluster\n", device_xname(sc->sc_dev));
		goto init_failure;
	}

	/* reload mac address */
	memcpy(sc->sc_enaddr, CLLADDR(ifp->if_sadl), ETHER_ADDR_LEN);
	// XXX: notyet

	/* program multicast address */
	gse_setmulti(sc);

	mii_mediachg(&sc->sc_mii);

	/* update if_flags */
	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	/* update local copy of if_flags */
	sc->sc_if_flags = ifp->if_flags;

	/* start RX DMA */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_DMA_CTRL_REG,
	    GSE_RX_DMA_CTRL_RX_EN);

 init_failure:
	splx(s);
	return error;
}

static void
gse_start(struct ifnet *ifp)
{
	struct gse_softc *sc;
	struct mbuf *m;
	int npkt;

	sc = ifp->if_softc;

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d: ifp=%p, sc=%p\n", __func__, __LINE__, ifp, sc);
#endif

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	npkt = 0;
	for (npkt = 0; ; npkt++) {
		IFQ_POLL(&ifp->if_snd, m);
		if (m == NULL)
			break;

		if (sc->sc_tx_free <= GSE_TX_RING_MINIMAL_CNT) {
			/* no tx descriptor now... */
			ifp->if_flags |= IFF_OACTIVE;
#ifdef DEBUG_GSE
	printf("%s:%d: TX Queue full\n", __func__, __LINE__);
#endif
			break;
		}

		IFQ_DEQUEUE(&ifp->if_snd, m);

		if (gse_encap(sc, m) != 0) {
			/* mbuf has too many chains */
			ifp->if_flags |= IFF_OACTIVE;
#ifdef DEBUG_GSE
			printf("%s:%d: TX Queue overflow. dropping packet\n", __func__, __LINE__);
#endif
			m_freem(m);
			ifp->if_oerrors++;
			break;
		}

#if NBPFILTER > 0
		/*
		 * If there's a BPF listener, bounce a copy of this frame
		 * to him.
		 */
		if (ifp->if_bpf != NULL)
			bpf_mtap(ifp->if_bpf, m);
#endif
	}

	if (npkt) {
		bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TX_DMA_CTRL_REG,
		    GSE_TX_DMA_CTRL_TX_EN);

		ifp->if_timer = 5;
	}
}

static void
gse_stop(struct ifnet *ifp, int disable)
{
	struct gse_softc *sc;
	int s;

	sc = ifp->if_softc;

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d: ifp=%p, disable=%d(do drain)\n", __func__, __LINE__, ifp, disable);
#endif

	s = splnet();

	/* stop DMA transfer */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TX_DMA_CTRL_REG, 0);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_DMA_CTRL_REG, 0);

	callout_stop(&sc->sc_rx_ch);

	/* Mark the interface as down and cancel the watchdog timer. */
	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);
	ifp->if_timer = 0;

	if (disable) {
		gse_drain_txbuf(sc);
		gse_drain_rxbuf(sc);
	}

	splx(s);

}

static void
gse_watchdog(struct ifnet *ifp)
{
	struct gse_softc *sc;
	int s;
	uint32_t swrst_bit;

	sc = ifp->if_softc;
	s = splnet();

	printf("%s: watchdog timeout\n", device_xname(sc->sc_dev));
	ifp->if_oerrors++;

	gse_tx_intr(sc);
	gse_rx_intr(sc);

	gse_stop(ifp, 1);

	/* force power off NIC */
	swrst_bit = STAR_REG_READ32(ORION_CLKPWR_SOFTRST_REG);
	swrst_bit &= ~ORION_CLKPWR_SOFTRST_SW;
	STAR_REG_WRITE32(ORION_CLKPWR_SOFTRST_REG, swrst_bit);
	gse_init(ifp);

	splx(s);
}

static int
gse_ioctl(struct ifnet *ifp, u_long command, void *data)
{
	struct gse_softc *sc;
	struct ifreq *ifr;
	int change, s, error;

	sc = ifp->if_softc;
	ifr = data;

	error = 0;
	s = splnet();

	switch (command) {
	case SIOCSIFFLAGS:
		if ((error = ifioctl_common(ifp, command, data)) != 0)
			break;

		change = ifp->if_flags ^ sc->sc_if_flags;

		/* up or down */
		if (change & IFF_UP) {
			if (ifp->if_flags & IFF_UP) {
#ifdef DEBUG_GSE_DUMP
				printf("%s:%d: NOW %s: DOWN -> UP\n", __func__, __LINE__, ifp->if_flags & IFF_RUNNING ? "RUNNING" : "STOPPED");
#endif
				/* down -> up */
				if ((ifp->if_flags & IFF_RUNNING) == 0)
					gse_init(ifp);
			} else {
#ifdef DEBUG_GSE_DUMP
				printf("%s:%d: NOW %s: UP -> DOWN\n", __func__, __LINE__, ifp->if_flags & IFF_RUNNING ? "RUNNING" : "STOPPED");
#endif
				/* up -> down */
				if ((ifp->if_flags & IFF_RUNNING) != 0)
					gse_stop(ifp, 1);
			}
		}

		/* into promisc mode or not */
		if ((change & (IFF_PROMISC | IFF_ALLMULTI)) != 0) {
			if ((ifp->if_flags & (IFF_PROMISC | IFF_ALLMULTI)) != 0) {
				GSE_REG_SETBIT(sc, GSE_SW_CFG_REG, GSE_SW_CFG_SKIP_L2_PORT0);
				GSE_REG_SETBIT(sc, GSE_SW_CFG_REG, GSE_SW_CFG_SKIP_L2_PORT1);
			} else {
				GSE_REG_CLRBIT(sc, GSE_SW_CFG_REG, GSE_SW_CFG_SKIP_L2_PORT0);
				GSE_REG_CLRBIT(sc, GSE_SW_CFG_REG, GSE_SW_CFG_SKIP_L2_PORT1);
			}
			gse_setmulti(sc);
		}

		sc->sc_if_flags = ifp->if_flags;
		error = 0;
		break;
	case SIOCSIFMEDIA:
	case SIOCGIFMEDIA:
#if notyet
		/* Flow control requires full-duplex mode. */
		if (IFM_SUBTYPE(ifr->ifr_media) == IFM_AUTO ||
		    (ifr->ifr_media & IFM_FDX) == 0)
			ifr->ifr_media &= ~IFM_ETH_FMASK;
		if (IFM_SUBTYPE(ifr->ifr_media) != IFM_AUTO) {
			if ((ifr->ifr_media & IFM_ETH_FMASK) == IFM_FLOW) {
				/* We can do both TXPAUSE and RXPAUSE. */
				ifr->ifr_media |=
				    IFM_ETH_TXPAUSE | IFM_ETH_RXPAUSE;
			}
			sc->sc_flowflags = ifr->ifr_media & IFM_ETH_FMASK;
		}
#endif
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_mii.mii_media, command);
		break;
	default:
		error = ether_ioctl(ifp, command, data);
		if (error != ENETRESET)
			break;

		error = 0;

		if (command == SIOCSIFCAP)
			error = (*ifp->if_init)(ifp);
		else if (command != SIOCADDMULTI && command != SIOCDELMULTI)
			;
		else if (ifp->if_flags & IFF_RUNNING) {
			/*
			 * Multicast list has changed; set the hardware filter
			 * accordingly.
			 */
			gse_setmulti(sc);
		}
		break;
	}

	splx(s);

	return error;
}

/*
 * for MII
 */
static int
gse_miibus_readreg(device_t dev, int phy, int reg)
{
	struct gse_softc *sc;
	struct ifnet *ifp;
	int timeout;
	uint32_t val, status;

	sc = device_private(dev);
	ifp = &sc->sc_ethercom.ec_if;

	/* read command */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_PHY_CTRL_REG,
	    GSE_PHY_CTRL_RW_OK | GSE_PHY_CTRL_RD_CMD |
	    GSE_PHY_CTRL_PHY_REG(reg) | GSE_PHY_CTRL_PHY_ADDR(phy));

	for (timeout = 5000; timeout > 0; --timeout) {
		status = bus_space_read_4(sc->sc_iot, sc->sc_ioh, GSE_PHY_CTRL_REG);
		if (status & GSE_PHY_CTRL_RW_OK)
			break;
	}
	if (timeout <= 0) {
		aprint_error_ifnet(ifp, "MII read timeout: reg=0x%02x\n", reg);
		val = -1;
	} else {
		val = (status >> 16) & 0xffff;
	}

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d: phy=%d, reg=%d => val=%08x (status=%08x)\n", __func__, __LINE__, phy, reg, val, status);
#endif

	return val;
}

static void
gse_miibus_writereg(device_t dev, int phy, int reg, int val)
{
	struct gse_softc *sc;
	struct ifnet *ifp;
	int timeout;
	uint32_t status;

	sc = device_private(dev);
	ifp = &sc->sc_ethercom.ec_if;

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d: phy=%d, reg=%d, val=0x%08x\n", __func__, __LINE__, phy, reg, val);
#endif

	/* write command */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_PHY_CTRL_REG,
	    GSE_PHY_CTRL_RW_OK | GSE_PHY_CTRL_WT_CMD |
	    GSE_PHY_CTRL_PHY_REG(reg) | GSE_PHY_CTRL_PHY_ADDR(phy) |
	    ((val & 0xffff) << 16));

	for (timeout = 5000; timeout > 0; --timeout) {
		status = bus_space_read_4(sc->sc_iot, sc->sc_ioh, GSE_PHY_CTRL_REG);
		if (status & GSE_PHY_CTRL_RW_OK)
			break;
	}
	if (timeout <= 0)
		aprint_error_ifnet(ifp, "MII write timeout: reg=0x%02x\n", reg);
}

static void
gse_miibus_statchg(device_t dev)
{
}

/*
 * misc routines
 */
static void
gse_init_txring(struct gse_softc *sc)
{
	int i;

	sc->sc_tx_free = GSE_TX_RING_CNT;
	sc->sc_tx_considx = 0;
	sc->sc_tx_prodidx = 0;

	/*
	 * build TX ring
	 */
	for (i = 0; i < GSE_TX_RING_CNT; i++) {
		sc->sc_txdesc_ring[i].tx_sdp = 0;
		sc->sc_txdesc_ring[i].tx_ctrl = 
		    GSE_TXDESC_CTRL_COWN;
		sc->sc_txdesc_ring[i].tx_vlan = 0;
		sc->sc_txdesc_ring[i].tx_rsvd = 0;
		TXDESC_WRITEOUT(i);
	}

	/* last descriptor */
	sc->sc_txdesc_ring[GSE_TX_RING_CNT - 1].tx_ctrl |=
	    GSE_TXDESC_CTRL_EOR;
	TXDESC_WRITEOUT(GSE_TX_RING_CNT - 1);
}

static int
gse_init_rxring(struct gse_softc *sc)
{
	int i, error;

	/*
	 * build RX ring
	 */
	for (i = 0; i < GSE_RX_RING_CNT; i++) {
		error = gse_alloc_rxbuf(sc, i);
		if (error != 0)
			return error;
		gse_reset_rxdesc(sc, i);
	}

	sc->sc_rx_readidx = 0;

	return 0;
}

static void
gse_reset_rxdesc(struct gse_softc *sc, int idx)
{
	sc->sc_rxdesc_ring[idx].rx_sdp =
	    sc->sc_rxsoft[idx].rxs_dmamap->dm_segs[0].ds_addr;
	sc->sc_rxdesc_ring[idx].rx_rsvd1 = 0;
	sc->sc_rxdesc_ring[idx].rx_rsvd2 = 0;
	RXDESC_WRITEOUT(idx);
	sc->sc_rxdesc_ring[idx].rx_ctrl =
	    ((idx == GSE_RX_RING_CNT - 1) ? GSE_RXDESC_CTRL_EOR : 0) |
	    GSE_RXDESC_CTRL_SDL(sc->sc_rxsoft[idx].rxs_dmamap->dm_mapsize);
	RXDESC_WRITEOUT(idx);
}

static int
gse_alloc_rxbuf(struct gse_softc *sc, int idx)
{
	struct mbuf *m;
	int error;

	KASSERT((idx >= 0) && (idx < GSE_RX_RING_CNT));

	/* free mbuf if already allocated */
	if (sc->sc_rxsoft[idx].rxs_mbuf != NULL) {
		bus_dmamap_unload(sc->sc_dmat, sc->sc_rxsoft[idx].rxs_dmamap);
		m_freem(sc->sc_rxsoft[idx].rxs_mbuf);
		sc->sc_rxsoft[idx].rxs_mbuf = NULL;
	}

	/* allocate new mbuf cluster */
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return ENOBUFS;
	MCLGET(m, M_DONTWAIT);
	if (!(m->m_flags & M_EXT)) {
		m_freem(m);
		return ENOBUFS;
	}
	sc->sc_rxsoft[idx].rxs_mbuf = m;

	error = bus_dmamap_load(sc->sc_dmat, sc->sc_rxsoft[idx].rxs_dmamap,
	    m->m_ext.ext_buf, m->m_ext.ext_size, NULL,
	    BUS_DMA_READ | BUS_DMA_NOWAIT);

	if (error)
		return error;

	gse_reset_rxdesc(sc, idx);
	return 0;
}

static void
gse_drain_txbuf(struct gse_softc *sc)
{
	int idx;
	struct gse_txsoft *txs;
	struct ifnet *ifp;

	ifp = &sc->sc_ethercom.ec_if;

	for (idx = sc->sc_tx_considx; idx != sc->sc_tx_prodidx;
	    idx = GSE_TX_NEXTIDX(idx)) {

		/* txsoft is used only first segment */
		txs = &sc->sc_txsoft[idx];
		TXDESC_READIN(idx);
		if (sc->sc_txdesc_ring[idx].tx_ctrl & GSE_TXDESC_CTRL_FS) {
			bus_dmamap_unload(sc->sc_dmat,
			    txs->txs_dmamap);
			m_freem(txs->txs_mbuf);

			ifp->if_oerrors++;
		}
		sc->sc_tx_free++;
	}
}

static void
gse_drain_rxbuf(struct gse_softc *sc)
{
	int i;

#ifdef DEBUG_GSE_DUMP
	printf("%s:%d\n", __func__, __LINE__);
#endif

	for (i = 0; i < GSE_RX_RING_CNT; i++) {
		if (sc->sc_rxsoft[i].rxs_mbuf != NULL) {
			bus_dmamap_unload(sc->sc_dmat, sc->sc_rxsoft[i].rxs_dmamap);
			m_freem(sc->sc_rxsoft[i].rxs_mbuf);
			sc->sc_rxsoft[i].rxs_mbuf = NULL;
		}
	}
}

static int
gse_alloc_ring(struct gse_softc *sc)
{
	int i, error;

	/*
	 * build DMA maps for TX.
	 * TX descriptor must be able to contain mbuf chains,
	 * So, make up GSE_MAX_PKT_NSEGS dmamap.
	 */
	for (i = 0; i < GSE_TX_RING_CNT; i++) {
		error = bus_dmamap_create(sc->sc_dmat, GSE_MAX_PKT_LEN,
		    GSE_MAX_PKT_NSEGS, GSE_MAX_PKT_LEN, 0, BUS_DMA_NOWAIT,
		    &sc->sc_txsoft[i].txs_dmamap);
		if (error) {
			aprint_error_dev(sc->sc_dev,
			    "can't create DMA map for TX descs\n");
			goto fail_1;
		}
	}

	/*
	 * build DMA maps for RX.
	 * RX descripter contains An mbuf cluster,
	 * and make up a dmamap.
	 */
	for (i = 0; i < GSE_RX_RING_CNT; i++) {
		error = bus_dmamap_create(sc->sc_dmat, MCLBYTES,
		    1, MCLBYTES, 0, BUS_DMA_NOWAIT,
		    &sc->sc_rxsoft[i].rxs_dmamap);
		if (error) {
			aprint_error_dev(sc->sc_dev,
			    "can't create DMA map for RX descs\n");
			goto fail_2;
		}
	}

	if (gse_alloc_dma(sc, sizeof(struct gse_txdesc) * GSE_TX_RING_CNT,
	    (void **)&(sc->sc_txdesc_ring), &(sc->sc_txdesc_dmamap)) != 0)
		return -1;
	memset(sc->sc_txdesc_ring, 0, sizeof(struct gse_txdesc) * GSE_TX_RING_CNT);

	if (gse_alloc_dma(sc, sizeof(struct gse_rxdesc) * GSE_RX_RING_CNT,
	    (void **)&(sc->sc_rxdesc_ring), &(sc->sc_rxdesc_dmamap)) != 0)
		return -1;
	memset(sc->sc_rxdesc_ring, 0, sizeof(struct gse_rxdesc) * GSE_RX_RING_CNT);

	if (gse_alloc_dma(sc, GSE_ETHER_MIN_LEN_VLAN,
	    (void **)&(sc->sc_padseg), &(sc->sc_padseg_dmamap)) != 0)
		return -1;
	memset(sc->sc_padseg, 0, GSE_ETHER_MIN_LEN_VLAN);

	return 0;

 fail_2:
	for (i = 0; i < GSE_RX_RING_CNT; i++) {
		if (sc->sc_rxsoft[i].rxs_dmamap != NULL)
			bus_dmamap_destroy(sc->sc_dmat,
			    sc->sc_rxsoft[i].rxs_dmamap);
	}
 fail_1:
	for (i = 0; i < GSE_TX_RING_CNT; i++) {
		if (sc->sc_txsoft[i].txs_dmamap != NULL)
			bus_dmamap_destroy(sc->sc_dmat,
			    sc->sc_txsoft[i].txs_dmamap);
	}
	return error;
}

static int
gse_encap(struct gse_softc *sc, struct mbuf *m)
{
	int csumflags, error, i, idx, padsize, port;
	bus_dmamap_t map;
	uint32_t ctrl;

	port = 0;	/* port0 */

	idx = sc->sc_tx_prodidx;
	map = sc->sc_txsoft[idx].txs_dmamap;

	csumflags = m->m_pkthdr.csum_flags;

	error = bus_dmamap_load_mbuf(sc->sc_dmat, map, m,
	    BUS_DMA_NOWAIT);
	if (error != 0) {
		aprint_error_dev(sc->sc_dev,
		    "Error mapping mbuf into TX chain: error=%d\n", error);
		m_freem(m);
		return error;
	}

	if (map->dm_nsegs > sc->sc_tx_free - GSE_TX_RING_MINIMAL_CNT) {
		bus_dmamap_unload(sc->sc_dmat, map);
		return ENOBUFS;
	}

#ifdef DEBUG_GSE_DUMP
	{
		uint8_t *p;
		int len = m_length(m);
		struct mbuf *tmp_m= m;

		printf("TX_INTR: m_flags=%08x, length=%d: ", m->m_flags, len);
		while (tmp_m != NULL) {
			printf("<paddr=0x%08x ", (uint32_t)(tmp_m->m_paddr));
			printf("len=%d>", tmp_m->m_len);

			p = mtod(tmp_m, uint8_t *);
			len = tmp_m->m_len;
			for (i = 0; i < len; i++) {
				printf(" %02x", *p++);
			}
			tmp_m = tmp_m->m_next;
		}
		printf("\n");
	}
#endif

//	padsize = GSE_ETHER_MIN_LEN_VLAN - m->m_pkthdr.len;
	padsize = GSE_ETHER_MIN_LEN - m->m_pkthdr.len;

	for (i = 0; i < map->dm_nsegs; i++) {
		ctrl = GSE_TXDESC_CTRL_PMAP(port);	/* to port0 */
		if (i == 0) {
			/* mark first segment */
			ctrl |= GSE_TXDESC_CTRL_FS;
			sc->sc_txsoft[idx].txs_mbuf = m;

			/* checksum offloading */
			if (csumflags & M_CSUM_IPv4)
				ctrl |= GSE_TXDESC_CTRL_ICO;
			if (csumflags & M_CSUM_UDPv4)
				ctrl |= GSE_TXDESC_CTRL_UCO;
			if (csumflags & M_CSUM_TCPv4)
				ctrl |= GSE_TXDESC_CTRL_TCO;

#ifdef DEBUG_GSE_DUMP
			if (csumflags & (M_CSUM_IPv4 | M_CSUM_UDPv4 | M_CSUM_TCPv4)) {
				printf("m=%p, csumflags=%s%s%s\n",
				    m,
				    (csumflags & M_CSUM_IPv4) ? "I" : "",
				    (csumflags & M_CSUM_UDPv4) ? "U" : "",
				    (csumflags & M_CSUM_TCPv4) ? "T" : "");
			}
#endif
		}
		if ((padsize <= 0) && (i == map->dm_nsegs - 1)) {
			/* mark last segment */
			ctrl |= GSE_TXDESC_CTRL_LS | GSE_TXDESC_CTRL_INT;
		}
		if (idx == GSE_TX_RING_CNT - 1) {
			/* mark end of ring */
			ctrl |= GSE_TXDESC_CTRL_EOR;
		}

		sc->sc_txdesc_ring[idx].tx_vlan = 0;
		sc->sc_txdesc_ring[idx].tx_sdp = map->dm_segs[i].ds_addr;
		TXDESC_WRITEOUT(idx);
		sc->sc_txdesc_ring[idx].tx_ctrl = ctrl |
		    GSE_TXDESC_CTRL_SDL(map->dm_segs[i].ds_len);
		TXDESC_WRITEOUT(idx);

#ifdef DEBUG_GSE_DUMP
		gse_dump_txdesc(sc, idx, "G_ENCAP         :");
#endif

		idx = GSE_TX_NEXTIDX(idx);
		sc->sc_tx_free--;
	}

	if (padsize > 0) {
		/* need padding */
		ctrl = GSE_TXDESC_CTRL_PMAP(port) |
		    GSE_TXDESC_CTRL_LS | GSE_TXDESC_CTRL_INT;
		if (idx == GSE_TX_RING_CNT - 1) {
			/* end of ring */
			ctrl |= GSE_TXDESC_CTRL_EOR;
		}
		sc->sc_txdesc_ring[idx].tx_vlan = 0;
		sc->sc_txdesc_ring[idx].tx_sdp = sc->sc_padseg_dmamap->dm_segs[0].ds_addr;
		TXDESC_WRITEOUT(idx);
		sc->sc_txdesc_ring[idx].tx_ctrl = ctrl |
		    GSE_TXDESC_CTRL_SDL(padsize);
		TXDESC_WRITEOUT(idx);
#ifdef DEBUG_GSE_DUMP
		gse_dump_txdesc(sc, idx, "G_ENCAP(padding):");
#endif

		idx = GSE_TX_NEXTIDX(idx);
		sc->sc_tx_free--;
	}

	bus_dmamap_sync(sc->sc_dmat, map, 0, map->dm_mapsize,
	    BUS_DMASYNC_PREWRITE);

#ifdef DEBUG_GSE_DUMP
	printf("tx_free=%d\n", sc->sc_tx_free);
#endif

	sc->sc_tx_prodidx = idx;

	return 0;
}


static int
gse_init_regs(struct gse_softc *sc)
{
	paddr_t paddr;
	uint32_t status;

	/* activate NIC */
	status = STAR_REG_READ32(ORION_CLKPWR_SOFTRST_REG);
	STAR_REG_WRITE32(ORION_CLKPWR_SOFTRST_REG, status |
	    ORION_CLKPWR_SOFTRST_SW);

	/* switch configuration */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SW_CFG_REG,
	    GSE_SW_CFG_NIC_MODE |
	    GSE_SW_CFG_CRC_STRIP |
	    GSE_SW_CFG_COL_MODE(3) |
	    GSE_SW_CFG_REV_MC_FILTER |
	    GSE_SW_CFG_BP_MODE(2) |
	    GSE_SW_CFG_BP_JAM_NO(10) |
	    GSE_SW_CFG_BKOFF_MODE(7) |
	    GSE_SW_CFG_HASH_ALG(0) |	/* direct mode */
	    GSE_SW_CFG_MAX_LEN(2) |
	    GSE_SW_CFG_AGE_TIME(0));

	/* port0 configuration */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_MAC_PORT0_CFG_REG,
	    GSE_MAC_PORT_CFG_LEARN_DIS |
	    GSE_MAC_PORT_CFG_BP_EN |
	    GSE_MAC_PORT_CFG_TXC_CHECK_EN |
	    GSE_MAC_PORT_CFG_FORCE_FC_TX |
	    GSE_MAC_PORT_CFG_FORCE_FC_RX |
	    GSE_MAC_PORT_CFG_FORCE_DUPLEX |
	    GSE_MAC_PORT_CFG_FORCE_SPEED(2) |
	    GSE_MAC_PORT_CFG_AN_EN);

	/* port1 configuration */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_MAC_PORT1_CFG_REG,
	    GSE_MAC_PORT_CFG_LEARN_DIS |
	    GSE_MAC_PORT_CFG_BP_EN |
	    GSE_MAC_PORT_CFG_TXC_CHECK_EN |
	    GSE_MAC_PORT_CFG_FORCE_FC_TX |
	    GSE_MAC_PORT_CFG_FORCE_FC_RX |
	    GSE_MAC_PORT_CFG_FORCE_DUPLEX |
	    GSE_MAC_PORT_CFG_FORCE_SPEED(2) |
	    GSE_MAC_PORT_CFG_AN_EN);

	/* cpu port configuration */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_CPU_PORT_CFG_REG,
	    GSE_CPU_PORT_CFG_OFFSET_2B |
	    GSE_CPU_PORT_CFG_AGE_EN |
	    GSE_CPU_PORT_CFG_LEARN_DIS |
	    GSE_CPU_PORT_CFG_PORT_DIS |
	    0);

	/* auto polling phy address, etc */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TEST1_REG, 0);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_AUTOPOLL_PHYADDR_REG, 0x00000040);


	/* priority setup */
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_PRI_CTRL_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_UDP_PRI_REG, 0xd0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS0_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS1_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS2_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS3_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS4_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS5_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS6_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_IP_TOS7_PRI_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SCHED_CTRL_REG, 0);

	/* rate limit */
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RATE_LIMIT_CTRL_REG, 0);

	/* flow control */
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_FLOW_CTRL_GLOBAL_THR_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_FLOW_CTRL_PORT_THR_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SMART_FLOW_CTRL_REG, 0);

	/* Port VID */
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_PORT_VID_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_GID01, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_GID23, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_GID45, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_GID67, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_PORT_MAP_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_VLAN_TAG_PORT_TAG_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SESSION_ID01_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SESSION_ID23_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SESSION_ID45_REG, 0);
//	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_SESSION_ID67_REG, 0);

	/* set TEST Register */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TEST0_REG, 0x44000000);


	/* TX Descriptor Physical Address */
	paddr = sc->sc_txdesc_dmamap->dm_segs[0].ds_addr;
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TX_DPTR_REG, paddr);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_TX_BASE_ADDR_REG, paddr);

	/* RX Descriptor Physical Address */
	paddr = sc->sc_rxdesc_dmamap->dm_segs[0].ds_addr;
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_DPTR_REG, paddr);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_RX_BASE_ADDR_REG, paddr);

	/* sync cache */
	bus_dmamap_sync(sc->sc_dmat, sc->sc_txdesc_dmamap, 0,
	    sc->sc_txdesc_dmamap->dm_mapsize, BUS_DMASYNC_PREWRITE);
	bus_dmamap_sync(sc->sc_dmat, sc->sc_rxdesc_dmamap, 0,
	    sc->sc_rxdesc_dmamap->dm_mapsize, BUS_DMASYNC_PREWRITE);

	/* Interrupts mode, etc */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh, GSE_DLY_INT_CFG_REG,
	 /* GSE_DLY_INT_CFG_MAX_PEND_TIME(10) | */
	 /* GSE_DLY_INT_CFG_MAX_INT_CNT(100) | */
	 /* GSE_DLY_INT_CFG_DELAY_INT_EN | */
	    0);

#ifdef DEBUG_GSE
	(void)gse_dump_txdesc;
	(void)gse_dump_rxdesc;
#endif

	return 0;
}

static int
gse_alloc_dma(struct gse_softc *sc, size_t size, void **addrp,
              bus_dmamap_t *mapp)
{
	bus_dma_segment_t seglist[1];
	int nsegs, error;

	if ((error = bus_dmamem_alloc(sc->sc_dmat, size, PAGE_SIZE, 0, seglist,
	    1, &nsegs, M_WAITOK)) != 0) {
		aprint_error_dev(sc->sc_dev,
		    "unable to allocate DMA buffer, error=%d\n", error);
		goto fail_alloc;
	}

	if ((error = bus_dmamem_map(sc->sc_dmat, seglist, 1, size, addrp,
	    BUS_DMA_NOWAIT | BUS_DMA_COHERENT)) != 0) {
		aprint_error_dev(sc->sc_dev,
		    "unable to map DMA buffer, error=%d\n",
		    error);
		goto fail_map;
	}

	if ((error = bus_dmamap_create(sc->sc_dmat, size, 1, size, 0,
	    BUS_DMA_NOWAIT, mapp)) != 0) {
		aprint_error_dev(sc->sc_dev,
		    "unable to create DMA map, error=%d\n", error);
		goto fail_create;
	}

	if ((error = bus_dmamap_load(sc->sc_dmat, *mapp, *addrp, size, NULL,
	    BUS_DMA_NOWAIT)) != 0) {
		aprint_error_dev(sc->sc_dev,
		    "unable to load DMA map, error=%d\n", error);
		goto fail_load;
	}

	return 0;

 fail_load:
	bus_dmamap_destroy(sc->sc_dmat, *mapp);
 fail_create:
	bus_dmamem_unmap(sc->sc_dmat, *addrp, size);
 fail_map:
	bus_dmamem_free(sc->sc_dmat, seglist, 1);
 fail_alloc:
	return error;
}


#ifdef DEBUG_GSE
static void
gse_dump_txdesc(struct gse_softc *sc, int idx, const char *prefix)
{
	char tmpbuf[128];

	TXDESC_READIN(idx);

	snprintb(tmpbuf, sizeof(tmpbuf), 
	    "\20"
	    "\40COWN"
	    "\37EOR"
	    "\36FS"
	    "\35LS"
	    "\34INT"
	    "\33FP"
	    "\32PRI2"
	    "\31PRI1"
	    "\30PRI0"
	    "\27FR"
	    "\26PMAP2"
	    "\25PMAP1"
	    "\24PMAP0"
	    "\23ICO"
	    "\22UCO"
	    "\21TCO",
	    sc->sc_txdesc_ring[idx].tx_ctrl);

	printf("%stxdesc[%3d]:tx_sdp =0x%08x\n", prefix, idx, sc->sc_txdesc_ring[idx].tx_sdp);
	printf("%stxdesc[%3d]:tx_ctrl=%s\n", prefix, idx, tmpbuf);
}


static void
gse_dump_rxdesc(struct gse_softc *sc, int idx, const char *prefix)
{
	char tmpbuf[128];

	RXDESC_READIN(idx);

	snprintb(tmpbuf, sizeof(tmpbuf), 
	    "\20"
	    "\40COWN"
	    "\37EOR"
	    "\36FS"
	    "\35LS"
	    "\34SP1"
	    "\33SP0"
	    "\32HR5"
	    "\31HR4"
	    "\30HR3"
	    "\27HR2"
	    "\26HR1"
	    "\25HR0"
	    "\24PROT1"
	    "\23PROT0"
	    "\22IPF"
	    "\21L4F",
	    sc->sc_rxdesc_ring[idx].rx_ctrl);

	printf("%srxdesc[%3d]:rx_sdp =0x%08x\n", prefix, idx, sc->sc_rxdesc_ring[idx].rx_sdp);
	printf("%srxdesc[%3d]:rx_ctrl=%s\n", prefix, idx, tmpbuf);

#if 0 /* HEXDUMP */
	{
		unsigned char *p;
		int i;

		p = mtod(sc->sc_rxsoft[idx].rxs_mbuf, char *);
		for (i = 0; i < 128; i++) {
			if ((i & 15) == 0)
				printf("%04x:", i);

			printf(" %02x", p[i]);

			if ((i & 15) == 15)
				printf("\n");
		}
	}
#endif
}
#endif
